import argparse
import atexit
import json
import os
import signal
import sys
import time
from collections import defaultdict
from multiprocessing import Event, Queue
from os import PathLike
from pathlib import Path
from pprint import pprint
from typing import Dict

from skyplane.broadcast.gateway.chunk_store import ChunkStore
from skyplane.broadcast.gateway.gateway_daemon_api import GatewayDaemonAPI
from skyplane.broadcast.gateway.gateway_queue import GatewayANDQueue, GatewayORQueue
from skyplane.broadcast.gateway.operators.gateway_operator import (
    GatewayWaitReciever,
    GatewayObjStoreReadOperator,
    GatewayRandomDataGen,
    GatewaySender,
    GatewayObjStoreWriteOperator,
    GatewayWriteLocal,
)
from skyplane.broadcast.gateway.operators.gateway_receiver import GatewayReceiver
from skyplane.utils import logger


# TODO: add default partition ID to main
# create gateway broadcast


class GatewayDaemon:
    def __init__(
        self,
        region: str,
        chunk_dir: PathLike,
        max_incoming_ports=64,
        use_tls=True,
        use_e2ee=False,
    ):
        # read gateway program
        gateway_program_path = Path(os.environ["GATEWAY_PROGRAM_FILE"]).expanduser()
        gateway_program = json.load(open(gateway_program_path, "r"))

        print(gateway_program)

        self.use_tls = use_tls

        # todo max_incoming_ports should be configurable rather than static
        self.region = region
        self.max_incoming_ports = max_incoming_ports

        # the chunk store managest the incoming queue of chunks and outgoing queue of chunk status updates
        self.chunk_store = ChunkStore(chunk_dir)

        self.error_event = Event()
        self.error_queue = Queue()
        if use_e2ee:
            e2ee_key_path = Path(os.environ["E2EE_KEY_FILE"]).expanduser()
            with open(e2ee_key_path, "rb") as f:
                self.e2ee_key_bytes = f.read()
        else:
            self.e2ee_key_bytes = None

        # create gateway operators
        self.terminal_operators = defaultdict(list)  # track terminal operators per partition
        self.operators = self.create_gateway_operators(gateway_program["_plan"])

        # single gateway reciever
        print("create gateway reciever")
        self.gateway_receiver = GatewayReceiver(
            "reciever",
            region=region,
            chunk_store=self.chunk_store,
            error_event=self.error_event,
            error_queue=self.error_queue,
            max_pending_chunks=max_incoming_ports,
            use_tls=self.use_tls,
            use_compression=False,  # use_compression,
            e2ee_key_bytes=self.e2ee_key_bytes,
        )

        # API server
        self.api_server = GatewayDaemonAPI(
            self.chunk_store, self.gateway_receiver, self.error_event, self.error_queue, self.terminal_operators
        )
        self.api_server.start()
        atexit.register(self.api_server.shutdown)
        logger.info(f"[gateway_daemon] API started at {self.api_server.url}")

    def create_gateway_operators(self, gateway_program: Dict):
        """Create a gateway plan from a gateway program"""

        operators = {}

        def create_output_queue(operator: Dict):
            # create output data queue
            print("DETERMINING OUTPUT QUEUE", operator["children"])
            if len(operator["children"]) == 0:
                return None
            if operator["children"][0]["op_type"] == "mux_and":
                return GatewayANDQueue()
            return GatewayORQueue()

        def get_child_operators(operator):
            if len(operator["children"]) == 0:
                return []
            if operator["children"][0]["op_type"] == "mux_and" or operator["children"][0]["op_type"] == "mux_or":
                return operator["children"][0]["children"]
            return operator["children"]

        def create_gateway_operators_helper(input_queue, program: Dict, partition_id: str):
            print("OPERATORS", operators)
            for op in program:

                handle = op["handle"]
                input_queue.register_handle(handle)
                print("INPUT QUEUE", input_queue, input_queue.get_handles())
                # create output data queue
                output_queue = create_output_queue(op)
                if output_queue is None:
                    # track what opeartors need to complete processing the chunk
                    self.terminal_operators[partition_id].append(op["handle"])

                # get child operators
                child_operators = get_child_operators(op)

                # create operators
                if op["op_type"] == "receive":
                    # wait for chunks from reciever
                    operators[handle] = GatewayWaitReciever(
                        handle=handle,
                        region=self.region,
                        input_queue=input_queue,
                        output_queue=output_queue,
                        n_processes=1,
                        chunk_store=self.chunk_store,
                        error_event=self.error_event,
                        error_queue=self.error_queue,
                    )
                elif op["op_type"] == "read_object_store":
                    operators[handle] = GatewayObjStoreReadOperator(
                        handle=handle,
                        region=self.region,
                        input_queue=input_queue,
                        output_queue=output_queue,
                        error_queue=self.error_queue,
                        error_event=self.error_event,
                        n_processes=op["num_connections"],
                        chunk_store=self.chunk_store,
                        bucket_name=op["bucket_name"],
                        bucket_region=op["bucket_region"],
                    )
                elif op["op_type"] == "gen_data":
                    operators[handle] = GatewayRandomDataGen(
                        handle=handle,
                        region=self.region,
                        input_queue=input_queue,
                        output_queue=output_queue,
                        error_queue=self.error_queue,
                        error_event=self.error_event,
                        chunk_store=self.chunk_store,
                        size_mb=op["size_mb"],
                    )
                elif op["op_type"] == "send":
                    operators[handle] = GatewaySender(
                        handle,
                        region=self.region,
                        ip_addr=op["ip_address"],
                        input_queue=input_queue,
                        output_queue=output_queue,
                        error_event=self.error_event,
                        error_queue=self.error_queue,
                        chunk_store=self.chunk_store,
                        use_tls=self.use_tls,
                        use_compression=False,  # operator["compress"],
                        e2ee_key_bytes=self.e2ee_key_bytes,
                        n_processes=op["num_connections"],
                    )
                elif op["op_type"] == "write_object_store":
                    operators[handle] = GatewayObjStoreWriteOperator(
                        handle=handle,
                        region=self.region,
                        input_queue=input_queue,
                        output_queue=output_queue,
                        error_queue=self.error_queue,
                        error_event=self.error_event,
                        n_processes=op["num_connections"],
                        chunk_store=self.chunk_store,
                        bucket_name=op["bucket_name"],
                        bucket_region=op["bucket_region"],
                    )
                elif op["op_type"] == "write_local":
                    operators[handle] = GatewayWriteLocal(
                        handle=handle,
                        region=self.region,
                        input_queue=input_queue,
                        output_queue=output_queue,
                        error_queue=self.error_queue,
                        error_event=self.error_event,
                        chunk_store=self.chunk_store,
                    )
                else:
                    raise ValueError(f"Unsupported op_type {op['op_type']}")
                # recursively create for child operators
                create_gateway_operators_helper(output_queue, child_operators, partition_id)

        print("GATEWAY PROGRAM")
        pprint(gateway_program)

        # create operator tree for each partition
        for partition, program in gateway_program.items():
            partition = str(partition)

            # create initial queue for partition
            self.chunk_store.add_partition(partition)

            create_gateway_operators_helper(
                self.chunk_store.chunk_requests[partition],  # incoming chunk requests for partition
                program,  # single partition program
                partition,
            )
        return operators

    def run(self):
        exit_flag = Event()

        def exit_handler(signum, frame):
            logger.warning("[gateway_daemon] Received signal {}. Exiting...".format(signum))
            exit_flag.set()
            for operator in self.operators.values():
                operator.stop_workers()
            sys.exit(0)

        for operator in self.operators.values():
            logger.info(f"[gateway_daemon] Starting gateway operator {operator.handle} workers")
            operator.start_workers()

        signal.signal(signal.SIGINT, exit_handler)
        signal.signal(signal.SIGTERM, exit_handler)

        logger.info("[gateway_daemon] Starting daemon loop")
        try:
            print(self.operators)
            while not exit_flag.is_set() and not self.error_event.is_set():

                print("pull queue...")
                self.api_server.pull_chunk_status_queue()
                # pull from chunk requests queue
                # print("running gateway daemon... nothing to do")
                # while True:
                #    try:
                #        chunk_req = self.chunk_store.chunk_requests.get_nowait()
                #    except Empty:
                #        break

                #    print("registered chunk", chunk_req.chunk.chunk_id, "partition", chunk_req.chunk.partition_id)
                #    partition_id = str(chunk_req.chunk.partition_id)
                #    if partition_id not in self.partition_queues:
                #        print(partition_id)
                #        print(list(self.partition_queues.keys()))
                #        raise ValueError(f"Partition {partition_id} does not exist in {list(self.partition_queues.keys())}")

                #    # queue the chunk if it needs to be
                #    if self.push_chunks[partition_id]:
                #        self.partition_queues[partition_id].put(chunk_req)

                #    print("listeners", self.partition_queues[partition_id].get_handles())
                #    for handle in self.partition_queues[partition_id].get_handles():
                #        print(self.operators[handle])

                # Check self.completed queue for chunks which have been processed by all operators
                # for partition, queue in self.completed.items():
                #    for chunk_req in queue.get_all():
                #        # unlink chunk that has finished processing
                #        chunk_file_path = self.chunk_store.get_chunk_file_path(chunk_req.chunk.chunk_id)
                #        chunk_file_path.unlink()
                #        self.chunk_store.state_finish_upload(chunk_req.chunk.chunk_id)
                #        logger.info(f"Finished processing chunk: {chunk_req.chunk.chunk_id}, partition: {chunk_req.chunk.partition_id}")

                time.sleep(0.1)  # yield

        except Exception as e:
            self.error_queue.put(e)
            self.error_event.set()
            logger.error(f"[gateway_daemon] Exception in daemon loop: {e}")
            logger.exception(e)

        # shut down workers except for API to report status
        logger.info("[gateway_daemon] Exiting all workers except for API")
        for operator in self.operators.values():
            operator.stop_workers()
        logger.info("[gateway_daemon] Done")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Skyplane Gateway Daemon")
    parser.add_argument("--region", type=str, required=True, help="Region tag (provider:region")
    parser.add_argument("--chunk-dir", type=Path, default="/tmp/skyplane/chunks", help="Directory to store chunks")
    parser.add_argument("--disable-tls", action="store_true")
    parser.add_argument("--use-compression", action="store_true")  # TODO: remove
    parser.add_argument("--disable-e2ee", action="store_true")  # TODO: remove
    args = parser.parse_args()

    os.makedirs(args.chunk_dir)
    daemon = GatewayDaemon(
        region=args.region,
        chunk_dir=args.chunk_dir,
        use_tls=not args.disable_tls,
    )
    daemon.run()



import functools
import os
from pathlib import Path


__config_root__ = Path("~/.skyplane").expanduser()
aws_config_path = __config_root__ / "aws_config"
azure_config_path = __config_root__ / "azure_config"
azure_sku_path = __config_root__ / "azure_sku_mapping"
gcp_config_path = __config_root__ / "gcp_config"


@functools.lru_cache(maxsize=None)
def load_config_path():
    if "SKYPLANE_CONFIG" in os.environ:
        path = Path(os.environ["SKYPLANE_CONFIG"]).expanduser()
    else:
        path = __config_root__ / "config"
    path.parent.mkdir(exist_ok=True)
    return path


@functools.lru_cache(maxsize=None)
def load_cloud_config(path):
    from skyplane.config import SkyplaneConfig

    if path.exists():
        return SkyplaneConfig.load_config(path)
    else:
        return SkyplaneConfig.default_config()


config_path = load_config_path()
cloud_config = load_cloud_config(config_path)
host_uuid_path = __config_root__ / "host_uuid"



from skyplane import compute
from skyplane.planner.topology import ReplicationTopology


class Planner:
    def __init__(self, src_provider: str, src_region, dst_provider: str, dst_region: str):
        self.src_provider = src_provider
        self.src_region = src_region
        self.dst_provider = dst_provider
        self.dst_region = dst_region

    def plan(self) -> ReplicationTopology:
        raise NotImplementedError


class DirectPlanner(Planner):
    def __init__(self, src_provider: str, src_region, dst_provider: str, dst_region: str, n_instances: int, n_connections: int):
        self.n_instances = n_instances
        self.n_connections = n_connections
        super().__init__(src_provider, src_region, dst_provider, dst_region)

    def plan(self) -> ReplicationTopology:
        src_region_tag = f"{self.src_provider}:{self.src_region}"
        dst_region_tag = f"{self.dst_provider}:{self.dst_region}"
        if src_region_tag == dst_region_tag:  # intra-region transfer w/o solver
            topo = ReplicationTopology()
            for i in range(self.n_instances):
                topo.add_objstore_instance_edge(src_region_tag, src_region_tag, i)
                topo.add_instance_objstore_edge(src_region_tag, i, src_region_tag)
            topo.cost_per_gb = 0
            return topo
        else:  # inter-region transfer w/ solver
            topo = ReplicationTopology()
            for i in range(self.n_instances):
                topo.add_objstore_instance_edge(src_region_tag, src_region_tag, i)
                topo.add_instance_instance_edge(src_region_tag, i, dst_region_tag, i, self.n_connections)
                topo.add_instance_objstore_edge(dst_region_tag, i, dst_region_tag)
            topo.cost_per_gb = compute.CloudProvider.get_transfer_cost(src_region_tag, dst_region_tag)
            return topo


class ILPSolverPlanner(Planner):
    def __init__(self, src_provider: str, src_region, dst_provider: str, dst_region: str, max_instances: int, max_connections: int):
        self.max_instances = max_instances
        self.max_connections = max_connections
        super().__init__(src_provider, src_region, dst_provider, dst_region)

    def plan(self) -> ReplicationTopology:
        raise NotImplementedError


class RONSolverPlanner(Planner):
    def __init__(self, src_provider: str, src_region, dst_provider: str, dst_region: str, max_instances: int, max_connections: int):
        self.max_instances = max_instances
        self.max_connections = max_connections
        super().__init__(src_provider, src_region, dst_provider, dst_region)

    def plan(self) -> ReplicationTopology:
        raise NotImplementedError



import base64
import datetime
import hashlib
import os
from functools import lru_cache
from xml.etree import ElementTree

import requests
from typing import Iterator, List, Optional, Tuple

from skyplane import exceptions, compute
from skyplane.config_paths import cloud_config
from skyplane.exceptions import NoSuchObjectException
from skyplane.obj_store.object_store_interface import ObjectStoreInterface, ObjectStoreObject
from skyplane.utils import logger


class GCSObject(ObjectStoreObject):
    def full_path(self):
        return os.path.join(f"gs://{self.bucket}", self.key)


class GCSInterface(ObjectStoreInterface):
    def __init__(self, bucket_name: str):
        self.bucket_name = bucket_name
        self.auth = compute.GCPAuthentication()
        self._gcs_client = self.auth.get_storage_client()
        self._requests_session = requests.Session()

    def path(self):
        return f"gs://{self.bucket_name}"

    @property
    @lru_cache(maxsize=1)
    def gcp_region(self):
        def map_region_to_zone(region) -> str:
            """Resolves bucket locations to a valid zone."""
            parsed_region = region.lower().split("-")
            if len(parsed_region) == 3:
                return region
            elif len(parsed_region) == 2 or len(parsed_region) == 1:
                # query the API to get the list of zones in the region and return the first one
                compute = self.auth.get_gcp_client()
                zones = compute.zones().list(project=self.auth.project_id).execute()
                for zone in zones["items"]:
                    if zone["name"].startswith(region):
                        return zone["name"]
            raise ValueError(f"No GCP zone found for region {region}")

        # load bucket from GCS client
        bucket = None
        default_region = cloud_config.get_flag("gcp_default_region")
        try:
            bucket = self._gcs_client.lookup_bucket(self.bucket_name)
        except Exception as e:
            # does not have storage.buckets.get access to the Google Cloud Storage bucket
            if "access to the Google Cloud Storage bucket" in str(e):
                logger.warning(
                    f"No access to the Google Cloud Storage bucket '{self.bucket_name}', assuming bucket is in the '{default_region}' zone"
                )
                return default_region
            raise
        if bucket is None:
            raise exceptions.MissingBucketException(f"GCS bucket {self.bucket_name} does not exist")
        else:
            loc = bucket.location.lower()
            if default_region.startswith(loc):
                loc = default_region
            return map_region_to_zone(loc)

    def region_tag(self):
        return "gcp:" + self.gcp_region

    def bucket(self) -> str:
        return self.bucket_name

    def bucket_exists(self):
        iterator = self._gcs_client.list_blobs(self.bucket_name, page_size=1)
        try:
            next(iterator.pages, None)
            return True
        except Exception as e:
            if "The specified bucket does not exist" in str(e):
                return False
            raise

    def exists(self, obj_name):
        try:
            self.get_obj_metadata(obj_name)
            return True
        except NoSuchObjectException:
            return False

    def create_bucket(self, gcp_region, premium_tier=True):
        assert premium_tier, "Standard tier GCS buckets are not supported"
        if not self.bucket_exists():
            bucket = self._gcs_client.bucket(self.bucket_name)
            bucket.storage_class = "STANDARD"
            region_without_zone = "-".join(gcp_region.split("-")[:2])
            self._gcs_client.create_bucket(bucket, location=region_without_zone)

    def delete_bucket(self):
        self._gcs_client.get_bucket(self.bucket_name).delete()

    def list_objects(self, prefix="") -> Iterator[GCSObject]:
        blobs = self._gcs_client.list_blobs(self.bucket_name, prefix=prefix)
        for blob in blobs:
            yield GCSObject("gcs", self.bucket_name, blob.name, blob.size, blob.updated, mime_type=getattr(blob, "content_type", None))

    def delete_objects(self, keys: List[str]):
        for key in keys:
            self._gcs_client.bucket(self.bucket_name).blob(key).delete()

    @lru_cache(maxsize=1024)
    def get_obj_metadata(self, obj_name):
        bucket = self._gcs_client.bucket(self.bucket_name)
        blob = bucket.get_blob(obj_name)
        if blob is None:
            raise NoSuchObjectException(
                f"Object {obj_name} does not exist in bucket {self.bucket_name}, or you do not have permission to access it"
            )
        return blob

    def get_obj_size(self, obj_name):
        return self.get_obj_metadata(obj_name).size

    def get_obj_last_modified(self, obj_name):
        return self.get_obj_metadata(obj_name).updated

    def get_obj_mime_type(self, obj_name):
        return self.get_obj_metadata(obj_name).content_type

    def send_xml_request(
        self,
        blob_name: str,
        params: dict,
        method: str,
        headers: Optional[dict] = None,
        expiration=datetime.timedelta(minutes=15),
        data=None,
        content_type="application/octet-stream",
    ):
        blob = self._gcs_client.bucket(self.bucket_name).blob(blob_name)

        headers = headers or {}
        headers["Content-Type"] = content_type

        # generate signed URL
        url = blob.generate_signed_url(
            version="v4", expiration=expiration, method=method, content_type=content_type, query_parameters=params, headers=headers
        )

        # prepare request
        if data:
            req = requests.Request(method, url, headers=headers, data=data)
        else:
            req = requests.Request(method, url, headers=headers)

        prepared = req.prepare()
        response = self._requests_session.send(prepared)

        if not response.ok:
            raise ValueError(f"Invalid status code {response.status_code}: {response.text}")

        return response

    def download_object(
        self, src_object_name, dst_file_path, offset_bytes=None, size_bytes=None, write_at_offset=False, generate_md5=False
    ) -> Tuple[Optional[str], Optional[bytes]]:
        src_object_name, dst_file_path = str(src_object_name), str(dst_file_path)
        src_object_name = src_object_name if src_object_name[0] != "/" else src_object_name
        bucket = self._gcs_client.bucket(self.bucket_name)
        blob = bucket.blob(src_object_name)

        # download object
        # TODO: download directly to file?
        if offset_bytes is None:
            chunk = blob.download_as_string()
        else:
            assert offset_bytes is not None and size_bytes is not None
            chunk = blob.download_as_string(start=offset_bytes, end=offset_bytes + size_bytes - 1)

        # write output
        if not os.path.exists(dst_file_path):
            open(dst_file_path, "a").close()
        if generate_md5:
            m = hashlib.md5()
        with open(dst_file_path, "wb+" if write_at_offset else "wb") as f:
            f.seek(offset_bytes if write_at_offset else 0)
            f.write(chunk)
            if generate_md5:
                m.update(chunk)
        md5 = m.digest() if generate_md5 else None
        mime_type = blob.content_type
        return mime_type, md5

    def upload_object(self, src_file_path, dst_object_name, part_number=None, upload_id=None, check_md5=None, mime_type=None):
        src_file_path, dst_object_name = str(src_file_path), str(dst_object_name)
        dst_object_name = dst_object_name if dst_object_name[0] != "/" else dst_object_name
        os.path.getsize(src_file_path)
        bucket = self._gcs_client.bucket(self.bucket_name)
        b64_md5sum = base64.b64encode(check_md5).decode("utf-8") if check_md5 else None

        if part_number is None:
            blob = bucket.blob(dst_object_name)
            blob.upload_from_filename(src_file_path, content_type=mime_type)
            if check_md5:
                blob_md5 = blob.md5_hash
                if b64_md5sum != blob_md5:
                    raise exceptions.ChecksumMismatchException(
                        f"Checksum mismatch for object {dst_object_name} in bucket {self.bucket_name}, "
                        + f"expected {b64_md5sum}, got {blob_md5}"
                    )
            return

        # multipart upload
        assert part_number is not None and upload_id is not None

        # send XML api request
        headers = {"Content-MD5": b64_md5sum} if check_md5 else None
        response = self.send_xml_request(
            dst_object_name,
            {"uploadId": upload_id, "partNumber": part_number},
            "PUT",
            headers=headers,
            data=open(src_file_path, "rb"),
        )

        # check response
        if "ETag" not in response.headers:
            raise exceptions.ChecksumMismatchException(
                f"Upload of object {dst_object_name} in bucket {self.bucket_name} failed, got status code {response.status_code} w/ response {response.text}"
            )

    def initiate_multipart_upload(self, dst_object_name: str, mime_type: Optional[str] = None) -> str:
        assert len(dst_object_name) > 0, f"Destination object name must be non-empty: '{dst_object_name}'"
        response = self.send_xml_request(dst_object_name, {"uploads": None}, "POST", content_type=mime_type)
        return ElementTree.fromstring(response.content)[2].text

    def complete_multipart_upload(self, dst_object_name, upload_id):
        # get parts
        response = self.send_xml_request(dst_object_name, {"uploadId": upload_id}, "GET")

        # build request xml tree
        tree = ElementTree.fromstring(response.content)
        ns = {"ns": tree.tag.split("}")[0][1:]}
        xml_data = ElementTree.Element("CompleteMultipartUpload")
        for part in tree.findall("ns:Part", ns):
            part_xml = ElementTree.Element("Part")
            etag_match = part.find("ns:ETag", ns)
            assert etag_match is not None
            etag = etag_match.text
            part_num_match = part.find("ns:PartNumber", ns)
            assert part_num_match is not None
            part_num = part_num_match.text
            ElementTree.SubElement(part_xml, "PartNumber").text = part_num
            ElementTree.SubElement(part_xml, "ETag").text = etag
            xml_data.append(part_xml)
        xml_data = ElementTree.tostring(xml_data, encoding="utf-8", method="xml")
        xml_data = xml_data.replace(b"ns0:", b"")

        try:
            # complete multipart upload
            response = self.send_xml_request(
                dst_object_name, {"uploadId": upload_id}, "POST", data=xml_data, content_type="application/xml"
            )
        except Exception as e:
            # cancel upload
            response = self.send_xml_request(dst_object_name, {"uploadId": upload_id}, "DELETE")
            raise exceptions.SkyplaneException("Failed to complete multipart upload") from e
